#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <string.h>
#include "CmMessage.h"
#include "FrameL.h"

#define FRAME_BUFFER_SIZE 10
char *Buffer[FRAME_BUFFER_SIZE];
int   NBytes[FRAME_BUFFER_SIZE];
int    FrGPS[FRAME_BUFFER_SIZE];
int   NFrames = 0;


/*------------------------------------------------------FdIOCmFrameHandler---*/
CmMessageStatus FdIOCmFrameHandler(CmMessage message, 
                                   char* sender,
                                   char* serverName)
/*---------------------------------------------------------------------------*/
{
  char *buf, *name; 
  int i, run, fnum, gtimeS, gtimeN, trig, nBytes;
   
  /*------------------------------------------------- decode Cm message -----*/
  name   = CmMessageGetText  (message);
  run    = CmMessageGetInt   (message);
  fnum   = CmMessageGetInt   (message);
  gtimeS = CmMessageGetInt   (message);
  gtimeN = CmMessageGetInt   (message);
  trig   = CmMessageGetInt   (message);
  nBytes = CmMessageGetInt   (message);
  buf    = CmMessageGetArray (message, CmMessageChar, &nBytes);

  if(NFrames < 10) printf("Receive frame %d from %s\n",gtimeS, sender);

  /*frame = FrameReadFromBuf (buf, *nBytes, 0);-*/

  /*--------------------------- store frame buffer in the circular buffer ---*/
  NFrames++;
  i = NFrames % FRAME_BUFFER_SIZE;

  if(NBytes[i] != 0) {
    printf("Error: overflow input frame buffer at GPS=%d\n",gtimeS);}
  else {
    Buffer[i] = malloc(nBytes);
    if(Buffer[i] == NULL) { 
      printf(" Malloc %d bytes buffer failed at GPS=%d\n",nBytes, gtimeS);}
    else {
      memcpy(Buffer[i], buf, nBytes);
      NBytes[i] = nBytes;
      FrGPS[i]  = gtimeS;
    }}

  return(CmMessageBreak);}

/*------------------------------------------------------FdIOCmFrameHandler---*/
char* FdGetFrameFromCm(double timeOut, int *nBytes, int *gps)
/*---------------------------------------------------------------------------*/
/* This function returns a pointer to a buffer containing the frame like if
   it was on a file. The size of the buffer is NBytes. 
   The buffer must be free by the user to avoid memory leak                  */
/*---------------------------------------------------------------------------*/
{
  static int init = 0;
  char *buffer;
  int i;

  if(init == 0) {
    CmMessageInstallHandler (FdIOCmFrameHandler, "FdFrame");
    init = 1;}

  CmMessageWaitWithTimeout(timeOut);

  i = NFrames % FRAME_BUFFER_SIZE;

  *nBytes = NBytes[i];
  *gps    = FrGPS[i];
  NBytes[i] = 0;
  buffer = Buffer[i];
  Buffer[i] = 0;

  return(buffer);
}

#if 0

/*----------------------------------------------------- FdPutFrameInMessage---*/
int FdPutFrameInMessage (CmMessage message, FrameH *frame, int compress)
/*----------------------------------------------------------------------------*/
{
  int nBytes;
  static char *buf;
  static int bufSize = 0;

  if (message == NULL) return(0);
  if (frame   == NULL) return(0);

  /*----------------------- Allocate the internal buffer for the first call---*/
  if (buf == NULL) {
    bufSize = 540000;
    buf     = (char *) malloc(bufSize);
    if (buf == NULL) return(0);}

  /*------Write the frame into the local buffer. If the buffer is too 
    small, increase its size (without limit) until the frame fits-------------*/
  do {
    nBytes = FrameWriteToBuf(frame, compress, buf, bufSize);

    if (nBytes == 0) {
      bufSize = 2 * bufSize;
      free(buf);
      buf = malloc(bufSize);
      if(buf == NULL) return(0);}
  } 
  while (nBytes == 0);

  /*------------------------------------------------- build the Cm message ---*/
  CmMessagePutText     (message, frame->name);
  CmMessagePutInt      (message, frame->run);
  CmMessagePutInt      (message, frame->frame);
  CmMessagePutInt      (message, frame->GTimeS);
  CmMessagePutInt      (message, frame->GTimeN);
  CmMessagePutInt      (message, 0);
  CmMessagePutInt      (message, nBytes);
  CmMessagePutExtArray (message, CmMessageChar, nBytes, buf);
  CmMessageSetType (message, "FdFrame");

  return(nBytes);
}

#endif

/*--------------------------------------------------------------FdSendFrame---*/
void FdSendFrame (char * cmName,
		  const char *destination,
		  struct FrameH *frame,
		  int debugPeriod)
/*----------------------------------------------------------------------------*/
{
  int status;
  CmMessage message;
  struct timeval tv;
  double currentTime, currentTime2;
  char   serData[256]; 
  static int lastGPS = 0, size = 0;
  static double sendingTime = 0., latency = 0.;

  if (frame == NULL) return;
 
  /*---------------------add statistic information about the previous frame---*/
  sprintf(serData,"LastGPS %d Latency %.1f Sending_Time %.1f nBytes %d",
	  lastGPS, latency, sendingTime, size); 
  FrSerDataNew(frame, cmName, frame->GTimeS, serData, 1./ frame->dt);

  /*-------------------------- send frame and measure time need to send it---*/
  gettimeofday(&tv, NULL);
  currentTime = tv.tv_sec + ( tv.tv_usec * 1.0e-6 );

  message = CmMessageNew ();
  size = FdPutFrameInMessage(message, frame, 9);
  status = CmMessageSend(message, (char *) destination);
  CmMessageDelete(message);

  if (status == 0) {
    printf( "Could not send frame GPS%d to %s latency=%.1f nBytes=%d\n", 
	    frame->GTimeS, destination, latency, size);
    FrameFree(frame);

    /*------- try to reconnect in case of problem---*/
    CmMessageCloseServer(cmName);
    CmConnectCleanup();
    if( !CmMessageOpenServer (cmName))
         {printf(" Cannot reopen Cm\n");}
    else {printf(" Reopen server with success\n");}

    return;}
 
  gettimeofday(&tv, NULL);
  currentTime2 = tv.tv_sec + ( tv.tv_usec * 1.0e-6 );
  sendingTime = currentTime2 - currentTime;
  latency = currentTime - FRGPSOFF - FRGPSTAI + frame->ULeapS - frame->GTimeS;
  if(frame->GTimeS % debugPeriod*frame->dt == 0)
    printf( "FdSendFrame>%d Latency=%.1f sending time=%.1f seconds size=%d\n", 
	    frame->GTimeS, latency, sendingTime, size);
  fflush(stdout);
  lastGPS = frame->GTimeS;

  FrameFree(frame);
 
  return;
}
